; Ansteuerung mehrerer LEDs an wenigen Pins eines AVR's
; Copyright 2004 Hagen Reddmann, mailto: HaReddmann AT T-Online DOT de
; Keine Haftungen in irgendeiner Weise, OpenSource, der Source darf frei verwendet werden
; in nicht kommerziellen Projekten, alles andere nach Absprache mit mir.
; Diese Copyright mu im Source verbleiben.
;
; Dieser Source demonstriert wie man an einem ATTiny15 mit nur 5 Pins insgesammt 20 LEDs
; individuell und mit 16 verschiedenen Helligkeiten ansteuern kann. In meinem Projekt 
; benutzte ich 10 SMT Duo LEDs die ber zwei Leitungen antiparallel geschaltet wurden.
; Bei 5 Pins werden die LEDs nur mit 20% Duty Cycle angesteuert.
; Weiter unten im Source findet man eine mgliche Beschaltung der LEDs.
; Die LEDs mssen in diesem Source an einem Port immer am Pin 0 bis X angeschlossen werden.
; Diese Restriktion lsst sich aber mit wenigen nderungen am Source vermeiden.
; Die LEDs werden immer so angesteuert das der Strom von Vcc nach Masse in den Pin fliet.
; LEDs die ausgeschaltet sind werden so angesteuert das das Datenrichtungsregister des Ports
; den Pin auf High-Z schaltet.
; Die Ansteuerung der LEDs wird in der Mainloop durch einen Mico-Interpreter erledigt. Mit Hilfe
; dieses Interpreters lassen sich ganz verschiedene Muster auf den LEDs ausgeben.
;
; Viel Spa, Hagen Reddmann


.nolist
.include "tn15def.inc"
.list


; Port und Pin Definitionen
.equ    LED_PORT        = PORTB
.equ    LED_DDR         = DDRB
.equ    LED_ROWS        = 5
.equ    LED_ROWS_MASK   = 0xFF << LED_ROWS

.if LED_ROWS > 5
  .error "This source supports not driving more as 5 pins without changes" 
.endif

.if LED_ROWS < 3
  .error "This source supports only driving of more as 2 pins"
.endif


; Registerfile
.def    isrSREG         = r1                    ; SREG Sicherung in ISR                    
.def    isrData         = r2                    ; Bitdaten der LEDs in ISR
.def    isrMapOffset    = r3                    ; Offset in das Mapping der LED Register
.def    isrLedRow       = r4                    ; aktuelle LED Row die auf 0=Low gesetzt wird


; LED Register, pro Register werden 2 LEDs gespeichert, 
; jeweils LED Rot/Grn in einem Nibble,
; ideal fr SMT Duo LEDs oder antiparallele Duo LEDs,
; somit kann jede Duo LED pro Farbe mit einer Helligkeit von 0 bis 15 individuell gesteuert werden.
.def    LED0            = r5
.def    LED1            = r6
.def    LED2            = r7

.if LED_ROWS > 3                                ; 4 Pins = 12 LEDs = 6 Duo Leds
  .def  LED3            = r8
  .def  LED4            = r9
  .def  LED5            = r10
.endif

.if LED_ROWS > 4                                ; 5 Pins = 20 LEDs = 10 Duo LEDs
  .def  LED6            = r11
  .def  LED7            = r12
  .def  LED8            = r13
  .def  LED9            = r14
.endif

.def    cmdWait         = r15                   ; Zeit bis zum nchsten Kommando

.def    cmdRepeat       = r20                   ; Wiederholungszhler fr eine Kommandoreihenfolge
.def    cmdStart        = r21                   ; Start/Stop Werte eines Kommandos
.def    cmdStop         = r22           
.def    cmdCount        = r23                   
.def    cmdDataLow      = r24                   ; Zeiger auf den aktuell zu interpretierenden Kommandocode
.def    cmdDataHigh     = r25

.def    Temp            = r26
.def    ledTime         = r27                   ; dient zur Synchronization 

.def    isrCounter      = r28                   ; berechnet Zeitpunkt wann die LED leuchten soll                          
.def    isrTemp         = r29                   ; Temporary fr ISR, wird nicht gesichert                



; Definitionen frs Mapping der LED's in das Registerfile
.equ    LED_g           = 0;                    ; Grn im Low Nibble
.equ    LED_r           = 1;                    ; Rot im High Nibble
.equ    LED_START       = 5;                    ; Register der LED0
.equ    LED_STOP        = 14;                   ; Register der LED9

.equ    LED_g0          =  5 * 2 + LED_g;       ; somit Register * 2 + Nibble
.equ    LED_r0          =  5 * 2 + LED_r; 
.equ    LED_g1          =  6 * 2 + LED_g;
.equ    LED_r1          =  6 * 2 + LED_r;
.equ    LED_g2          =  7 * 2 + LED_g;
.equ    LED_r2          =  7 * 2 + LED_r;


.if LED_ROWS > 3
  .equ  LED_g3          =  8 * 2 + LED_g;
  .equ  LED_r3          =  8 * 2 + LED_r; 
  .equ  LED_g4          =  9 * 2 + LED_g;
  .equ  LED_r4          =  9 * 2 + LED_r;
  .equ  LED_g5          = 10 * 2 + LED_g;
  .equ  LED_r5          = 10 * 2 + LED_r;
.endif

.if LED_ROWS > 4
  .equ  LED_g6          = 11 * 2 + LED_g;
  .equ  LED_r6          = 11 * 2 + LED_r; 
  .equ  LED_g7          = 12 * 2 + LED_g;
  .equ  LED_r7          = 12 * 2 + LED_r;
  .equ  LED_g8          = 13 * 2 + LED_g;
  .equ  LED_r8          = 13 * 2 + LED_r;
  .equ  LED_g9          = 14 * 2 + LED_g;
  .equ  LED_r9          = 14 * 2 + LED_r;
.endif


; Makros
.macro outi
        ldi     Temp, @1
        out     @0, Temp
.endmacro


; Code
.org 0
        rjmp    Reset
.org T0OVaddr
        rjmp    isrT0OV


; Mappings der LEDs von ihrer physikalischen Position am Port in das Registerfile
; Diese Mapping hat zwei Aufgaben
; 1.) es ermglicht die Zuordnung der Pins zu Registern 
; 2.) mit Hilfe diese Mappings kann man ein flexibles Hardwaredesign machen und die
;     Ansteuerung der LEDs dynamisch an die Register anpassen

ledRegisterMap:

; Beispiel fr 3 SMT Duo Leds Rot/Grn = 6 Leds an 3 Pins
; ergibt 33% Dutycycle pro LED.
; Row/Pin      2       1       0                # SMT Duo Led
;  0           |       |--|>|--|      LED_r0         0
;              |-------|--|>|--|      LED_r1         1
;              |       |       |       
;  1           |       |--|<|--|      LED_g0         0
;              |--|>|--|       |      LED_r2         2
;              |       |       |  
;  2           |--|<|--|-------|      LED_g1         1
;              |--|<|--|       |      LED_g2         2

; die Rows mssen revers in die Tabelle eingetragen werden da isrLEDRows rechtsge-shiftet wird.
.if LED_ROWS == 3
  .db          0, LED_g2, LED_g1,    LED_r2,      0, LED_g0
  .db     LED_r1, LED_r0,      0  
.endif


; Beispiel fr 6 SMT Duo Leds Rot/Grn = 12 Leds an 4 Pins
; ergibt 25% Dutycycle pro LED.
; Row/Pin      3       2       1       0                # SMT Duo Led
;  0           |       |       |--|>|--|      LED_r0         0
;              |       |-------|--|>|--|      LED_r1         1
;              |-------|-------|--|>|--|      LED_r2         2
;              |       |       |       |
;  1           |       |       |--|<|--|      LED_g0         0
;              |       |--|>|--|       |      LED_r3         3
;              |-------|--|>|--|       |      LED_r4         4
;              |       |       |       |
;  2           |       |--|<|--|-------|      LED_g1         1
;              |       |--|<|--|       |      LED_g3         3
;              |--|>|--|       |       |      LED_r5         5
;              |       |       |       |
;  3           |--|<|--|-------|-------|      LED_g2         2
;              |--|<|--|-------|       |      LED_g4         4
;              |--|<|--|       |       |      LED_g5         5       


.if LED_ROWS == 4
  .db          0, LED_g5, LED_g4, LED_g2     ; Row 3  isrLedRow = 8 = 0b1000  
  .db     LED_r5,      0, LED_g3, LED_g1     ; Row 2  isrLedRow = 4 = 0b0100
  .db     LED_r4, LED_r3,      0, LED_g0     ; Row 1  isrLedRow = 2 = 0b0010
  .db     LED_r2, LED_r1, LED_r0,      0     ; Row 0  isrLedRow = 1 = 0b0001
.endif

; 20 LEDs and 5 Pins, 20% Dutycycle
.if LED_ROWS == 5
  .db          0, LED_g9, LED_g8, LED_g6, LED_g3,    LED_r9,      0, LED_g7, LED_g5, LED_g2   
  .db     LED_r8, LED_r7,      0, LED_g4, LED_g1,    LED_r6, LED_r5, LED_r4,      0, LED_g0   
  .db     LED_r3, LED_r2, LED_r1, LED_r0,      0   
.endif

; 6 Pins = 30 LEDs, 17% Duty
; 7 Pins = 42 LEDs, 14% Duty 
; 8 Pins = 56 LEDs, 13% Duty


; Timer0 ISR zur Soft-PWM Ansteuerung der LEDs 
; benutzt Register r0, ZL, ZH immer 0, und alle isrXXXX Register
; Register r0, ZL werden NICHT gesichert !!
isrT0OV:        
        in      isrSREG, SREG
        
        lsr     isrLedRow                       ; schiebe LED Low-Pegel Row nach rechts
        brne    isr_1                           ; alle LEDs an diesem Pin = Anode werden angesteuert

        ldi     isrTemp, 1 << (LED_ROWS -1)     ; eine PWM Welle ist durch, beginne von neuem
        mov     isrLedRow, isrTemp
        ldi     isrTemp, Low(ledRegisterMap) * 2
        mov     isrMapOffset, isrTemp

        subi    isrCounter, Low(-7)             ; bestimmt wann eine LED leuchten mu auf Grund der
        andi    isrCounter, 0x0F                ; eingestellten Helligkeit
        brne    isr_1                           
                                              
        inc     ledTime                         ; nach jedem vollstndigen Durchlauf erhhen wir den
                                                ; Synchronizations Zhler
      
isr_1:  clr     isrData                         ; speichert das Bitmuster der anzusteurenden LEDs  
        ldi     isrTemp, LED_ROWS               ; Schleifenzhler 

isr_2:  mov     ZL, isrMapOffset                ; lade Register-Inhalt der LED ber das Mapping
        inc     isrMapOffset
        lpm
        lsr     r0      
        breq    isr_4                           ; wenn Register == 0 dann keine LED angeschlossen
        mov     ZL, r0
        ld      ZL, Z                           ; lade Inhalt == Helligkeit
        brcc    isr_3
        swap    ZL                              
isr_3:  andi    ZL, 0x0F
        cp      isrCounter, ZL                  ; soll LED auf Grund der Helligekt leuchten ??
isr_4:  rol     isrData
        dec     isrTemp
        brne    isr_2
                
        in      ZL, LED_DDR                     ; nun setze Port und Datenregister, wir
        andi    ZL, LED_ROWS_MASK               ; setzen entwerder 0,1 oder High-Z
        or      ZL, isrLedRow                   ; Actung! wird im restlichen Source LED_DDR oder
        or      ZL, isrData                     ; oder LED_PORT manipuliert so mu in diesem
                                                ; Source der Code in einem CLI/SEI Block ausgefhrt werden
        in      isrTemp, LED_PORT               ; Selbiges gilt fr den Zugriff auf r0,ZL,ZH
        andi    isrTemp, LED_ROWS_MASK    
        or      isrData, isrTemp

        out     LED_PORT, isrTemp
        out     LED_DDR, ZL
        out     LED_PORT, isrData

        out     SREG, isrSREG       
        reti



; Initialization
Reset:  outi    SPL, Low(RAMEND)
;        outi    SPH, High(RAMEND)
        outi    OSCCAL, 0x5A            ; Calibration, hardcoded, abhngig vom Chip
        
        clr     ZH                      ; unbedingt ntig, auch wenn Datenblatt was anderes sagt
        clr     isrLedRow
        clr     isrCounter

        outi    TCCR0, (0 << CS02) | (0 << CS01) | (1 << CS00) ; 
        outi    TIMSK, (1 << TOIE0)
        outi    MCUCR, (1 << SE)

; hier beginnt die Mainloop, sie liest aus Commands ein Kommando mit entsprechenden Parametern
; interpriertiert diese und fhrt die jeweilige Aktion auf die LEDs aus.
Start:  clt
        ldi     cmdDataLow, Low(Commands * 2)
        ldi     cmdDataHigh, High(Commands * 2)
        ldi     cmdRepeat, 0xFF
        ldi     Temp, 10
        mov     cmdWait, Temp

Finish: brtc    DoCmd
Main:   rcall   Wait
DoCmd:  rcall   Next

        mov     cmdCount, r0
        andi    cmdCount, 0x0F
        mov     Temp, r0
        bst     Temp, 7
        swap    Temp
        andi    Temp, 0x07
        brne    cmd_10
        ; commands 0x0?
cmd_00: tst     cmdCount
        breq    Finish
        dec     cmdCount
        brne    cmd_03
cmd_01: ; Clear
        ldi     ZL, LED_START
cmd_02: st      Z, cmdCount     ; cmdCount == 0
        inc     ZL
        cpi     ZL, LED_STOP +1
        brne    cmd_02
        rjmp    Finish
cmd_03: dec     cmdCount
        brne    cmd_06
       
        ; Repeat Commands, -Offset, Count
        rcall   Next
        mov     Temp, r0
        rcall   Next
        cpi     cmdRepeat, 0xFF
        brne    cmd_04
        mov     cmdRepeat, r0
        rjmp    cmd_05 
cmd_04: subi    cmdRepeat, 1
        brcs    Finish
cmd_05: subi    cmdDataLow, 3
        sbci    cmdDataHigh, 0
        sub     cmdDataLow, Temp
        sbci    cmdDataHigh, 0
        rjmp    Finish

cmd_06: ; other 0x0? Commands 0x02 upto 0x0F
        rjmp    Start

cmd_10: rcall   Next 
        cpi     Temp, 1
        brne    cmd_20
        ; Setup, Index, Value
        ldi     ZL, LED_START
        add     ZL, cmdCount
        st      Z, r0
        rjmp    Finish
cmd_20: mov     cmdStart, r0
        cpi     Temp, 4
        brsh    cmd_40
        ; Left or Right Rotate, Count, Start, Stop
        swap    cmdStart
        andi    cmdStart, 0x0F
        subi    cmdStart, Low(-LED_START)
        mov     cmdStop, r0
        andi    cmdStop, 0x0F
        subi    cmdStop, Low(-LED_START)
        cp      cmdStop, cmdStart
        brlo    Finish
        cpi     Temp, 2
        brne    cmd_30
        ; Left Rotate, Count, Start, Stop
cmd_21: mov     ZL, cmdStop
        ld      Temp, Z
        mov     ZL, cmdStart   
cmd_22: ld      r0, Z
        st      Z, Temp
        mov     Temp, r0
        inc     ZL      
        cp      cmdStop, ZL 
        brsh    cmd_22 
        brtc    cmd_23
        rcall   Wait
cmd_23: subi    cmdCount, 1
        brcc    cmd_21
        rjmp    DoCmd
cmd_30: ; Right Rotate, Count, Start, Stop
cmd_31: mov     ZL, cmdStart
        ld      Temp, Z
        mov     ZL, cmdStop
cmd_32: ld      r0, Z
        st      Z, Temp
        mov     Temp, r0
        dec     ZL
        cp      ZL, cmdStart
        brsh    cmd_32
        brtc    cmd_33
        rcall   Wait
cmd_33: subi    cmdCount, 1
        brcc    cmd_31
        rjmp    DoCmd
cmd_40: cpi     Temp, 5
        brsh    cmd_50
        ; Left Shift In, Count, Value
cmd_41: ldi     ZL, LED_START
        mov     Temp, cmdStart
cmd_42: ld      r0, Z
        st      Z, Temp
        mov     Temp, r0
        inc     ZL
        cpi     ZL, LED_STOP +1
        brne    cmd_42
        brtc    cmd_43
        rcall   Wait
cmd_43: subi    cmdCount, 1
        brcc    cmd_41
        rjmp    DoCmd
        ; Right Shift In, Count, Value
cmd_50: cpi     Temp, 6
        brsh    cmd_60
cmd_51: ldi     ZL, LED_STOP
        mov     Temp, cmdStart
cmd_52: ld      r0, Z
        st      Z, Temp
        mov     Temp, r0
        dec     ZL
        cpi     ZL, LED_START -1
        brne    cmd_52
        brtc    cmd_53
        rcall   Wait
cmd_53: subi    cmdCount, 1
        brcc    cmd_51
        rjmp    DoCmd
        ; Command 0x60, 0x70
cmd_60: ldi     ZL, LED_START
        add     ZL, cmdCount
        ld      r0, Z
        cpi     Temp, 6
        brne    cmd_70
        add     r0, cmdStart
        rjmp    cmd_71
cmd_70: sub     r0, cmdStart
cmd_71: st      Z, r0
        rjmp    Finish

; lade nchstes Kommando
Next:   mov     ZL, cmdDataLow 
        mov     ZH, cmdDataHigh
        lpm
        clr     ZH
        inc     cmdDataLow   
        brne    PC +2
        inc     cmdDataHigh
        ret

; warte auf Sync
Wait:   sei
        sleep
        cp      ledTime, cmdWait
        brlo    Wait
        cli
        clr     ledTime
        ret


; Kommandobytes in Bits  w ccc iiii
; w    = Wait Flag, falls 1 so mu nach diesem Kommando auf ledTime >= cmdWait gewartet werden
; ccc  = Kommandocode, 0-7
; iiii = zustzliche Daten, zB. Index in die Register von LED0 bis LED9 oder Register cmdWait


; Befehle:
; 
; ohne Warten,  mit Warten    Name                               Bemerkung
; 00            80,           NOP, Wait                          keine Operation, Wartezyklus
; 01            81            Clear                              setzt LED0-9 auf 0x00, alle LEDs aus
; 02, ss, cc    92, ss, cc    Repeat Commands, -Offset, Count    wiederholt vorhergehenden Komandblock, KEINE Verschachtelungen mglich !!
; 0F            8F            Restart, Terminator                beendet Codes und startet von neuem  
;
; 03            83 
; bis
; 0E            8E            unused                             unbenutzt
;
; 1i, vv        9i, vv        Setup, Index, Value                Setzt LED[i] auf Wert vv
; 2c, se        Ac, se        Left-Rotate, Count, Start, End     Links Rotiert LED[s] bis LED[e] c+1 mal
; 3c, se        Bc, se        Right-Rotate, Count, Start, End    Rechts Rotiert LED[s] bis LED[e] c+1 mal
; 4c, vv        Cc, vv        Left-Shift-in, Count, Value        c+1 mal Links Schieben der LED[], rechte Stelle LED[0] wird mit vv gefllt
; 5c, vv        Dc, vv        Right-Shift-in, Count, Value       c+2 mal Rechts Schieben der LED[], rechte Stelle LED[9] wird mit vv gefllt
; 6i, vv        Ei, vv        Increment, Index, Value            LED[i] = LED[i] + vv
; 7i, vv        Fi, vv        Decrement, Index, Value            LED[i] = LED[i] - vv 

; die Kommandos 1i,9i,6i,Ei,7i,Fi knnen auch benutzt werden um Register cmdWait = Wartezeit
; zu ndern. i mu 10 = 0x0A sein. Mit dem Befehl "0x1A, 0x0A" wird cmdWait = 10 gesetzt.

  
       

Commands:
.db  0x01        ; Clear
.db  0x1A, 0x03  ; cmdWait = 3 

; Flashing der LEDs, Helligkeiten inc/decrement
.db  0x64, 0x30, 0xE5, 0x03   
.db  0x64, 0x30, 0x65, 0x03, 0x63, 0x30, 0xE6, 0x03   
.db  0x64, 0x30, 0x65, 0x03, 0x63, 0x30, 0x66, 0x03, 0x62, 0x30, 0xE7, 0x03      
.db  0x64, 0x30, 0x65, 0x03, 0x63, 0x30, 0x66, 0x03, 0x62, 0x30, 0x67, 0x03, 0x61, 0x30, 0xE8, 0x03      
.db  0x64, 0x30, 0x65, 0x03, 0x63, 0x30, 0x66, 0x03, 0x62, 0x30, 0x67, 0x03, 0x61, 0x30, 0x68, 0x03, 0x60, 0x30, 0xE9, 0x03

.db  0x74, 0x30, 0x75, 0x03, 0x73, 0x30, 0x76, 0x03, 0x72, 0x30, 0x77, 0x03, 0x71, 0x30, 0x78, 0x03, 0x70, 0x30, 0xF9, 0x03
.db  0x74, 0x30, 0x75, 0x03, 0x73, 0x30, 0x76, 0x03, 0x72, 0x30, 0x77, 0x03, 0x71, 0x30, 0xF8, 0x03      
.db  0x74, 0x30, 0x75, 0x03, 0x73, 0x30, 0x76, 0x03, 0x72, 0x30, 0xF7, 0x03      
.db  0x74, 0x30, 0x75, 0x03, 0x73, 0x30, 0xF6, 0x03   
.db  0x74, 0x30, 0xF5, 0x03   
; wiederhole Flashing 10x
.db  0x02, 0x78, 0x0A

; cmdWait = 16
.db  0x1A, 0x10   

.db  0xC1, 0xF0, 0xC5, 0x00, 0xC1, 0x0F
.db  0x30, 0x58, 0xA0, 0x14, 0x02, 0x04, 0x01
.db  0x20, 0x58, 0xB0, 0x14, 0x02, 0x04, 0x01
.db  0x01   

; cmdWait = 6
.db  0x1A, 0x06   

.db  0x90, 0xF0, 0xC8, 0x00 
.db  0x90, 0x0F, 0xA7, 0x08
.db  0x90, 0x3F, 0xA6, 0x07 
.db  0xD6, 0x00

.db  0xA6, 0x29, 0xA6, 0x18, 0xA6, 0x07
.db  0xB6, 0x07, 0xB6, 0x18, 0xB6, 0x29
.db  0xA6, 0x29, 0x99, 0x00
.db  0xA7, 0x19, 0x99, 0x00
.db  0xC9, 0x00   

; cmdWait = 3
.db  0x1A, 0x03
    
.db  0x90, 0xF0, 0xC9, 0x00, 0x99, 0x0F, 0xD9, 0x00   
.db  0x90, 0x3F, 0xC9, 0x00, 0x99, 0xF0, 0xD9, 0x00   
.db  0x90, 0x0F, 0xC9, 0x00, 0x99, 0x3F, 0xD9, 0x00   

; cmdWait = 10
.db  0x1A, 0x0A

.db  0xC0, 0xF0, 0xC0, 0x0F, 0xC0, 0x3F, 
.db  0xAF, 0x09, 0xAF, 0x09, 0xAE, 0x09, 
.db  0xBE, 0x09, 0xBF, 0x09, 0xBF, 0x09, 
.db  0xD2, 0x00

; cmdWait = 20
.db  0x1A, 0x14

.db  0xC9, 0xF0, 0xC9, 0x0F, 0xC9, 0x3F, 0xC9, 0x00 
.db  0xD9, 0x3F, 0xD9, 0x0F, 0xD9, 0xF0, 0xD9, 0x00   

      
.db  0x0F   